use std::time;

use async_trait::async_trait;
use chrono::Utc;
use log::warn;
use rand::Rng;
use uuid::Uuid;

use crate::{
    message_handler::user_commands::{
        movement::attempt_move_immediate, CommandHandlingError, UResult,
    },
    models::{
        item::{Item, LocationActionType, SkillType},
        task::{Task, TaskDetails, TaskMeta},
    },
    regular_tasks::{
        queued_command::{MovementSource, QueueCommand, QueuedCommandContext},
        TaskHandler, TaskRunContext,
    },
    static_content::{
        possession_type::{possession_data, DamageType},
        room::{room_map_by_code, Direction, MaterialType, Room},
    },
    DResult,
};

use super::{combat::change_health, comms::broadcast_to_room, skills::skill_check_and_grind};

pub struct EffectEnvironmentHandler;
pub struct ConsequenceEnvironmentHandler;

pub async fn ensure_appropriate_environment_handler_after_movement(
    ctx: &mut QueuedCommandContext<'_>,
) -> UResult<()> {
    if ctx.item.item_type != "player" {
        return Ok(());
    }
    let (loc_type, loc_code) = match ctx.item.location.split_once("/") {
        None => return Ok(()),
        Some(v) => v,
    };
    let need_handler = if loc_type != "room" {
        false
    } else {
        let room = match room_map_by_code().get(loc_code) {
            None => return Ok(()),
            Some(r) => r,
        };
        let water_effects = match room.material_type {
            MaterialType::WaterSurface | MaterialType::Underwater => true,
            _ => false,
        };
        let env_effects = room.environment.radiation > 0 || room.environment.passive_health != 0;
        water_effects || env_effects
    };

    let existing_task = ctx
        .trans
        .fetch_specific_task("ApplyEnvironmentalEffects", &ctx.item.refstr())
        .await?;
    if need_handler {
        if existing_task.is_none() {
            ctx.trans
                .upsert_task(&Task {
                    meta: TaskMeta {
                        task_code: ctx.item.refstr(),
                        next_scheduled: Utc::now() + chrono::TimeDelta::try_seconds(10).unwrap(),
                        ..Default::default()
                    },
                    details: TaskDetails::ApplyEnvironmentalEffects {
                        on_player_type: ctx.item.item_type.clone(),
                        on_player_code: ctx.item.item_code.clone(),
                    },
                })
                .await?;
        }
    } else {
        if existing_task.is_some() {
            ctx.trans
                .delete_task("ApplyEnvironmentalEffects", &ctx.item.refstr())
                .await?;
        }
    }
    Ok(())
}

#[async_trait]
impl TaskHandler for EffectEnvironmentHandler {
    async fn do_task(&self, ctx: &mut TaskRunContext) -> DResult<Option<time::Duration>> {
        let (player_type, player_code) = match &ctx.task.details {
            TaskDetails::ApplyEnvironmentalEffects {
                on_player_type,
                on_player_code,
            } => (on_player_type, on_player_code),
            x => {
                warn!(
                    "EnvironmentHandler called with bad TaskDetail type: {:#?}",
                    x
                );
                return Ok(None);
            }
        };

        let player = match ctx
            .trans
            .find_item_by_type_code(&player_type, &player_code)
            .await?
        {
            None => return Ok(None),
            Some(p) => p,
        };
        let mut player = (*player).clone();

        let (loc_type, loc_code) = match player.location.split_once("/") {
            None => return Ok(None),
            Some(v) => v,
        };
        if loc_type != "room" {
            return Ok(None);
        };

        let room = match room_map_by_code().get(loc_code) {
            None => return Ok(None),
            Some(r) => r,
        };

        let need_save = water_environment_effects(ctx, &room, &mut player).await?
            || rad_environment_effects(ctx, &room, &mut player).await?
            || passive_health_environment_effects(ctx, &room, &mut player).await?;

        if need_save {
            ctx.trans.save_item_model(&player).await?;
        }

        Ok(Some(time::Duration::from_secs(10)))
    }
}

#[async_trait]
impl TaskHandler for ConsequenceEnvironmentHandler {
    async fn do_task(&self, ctx: &mut TaskRunContext) -> DResult<Option<time::Duration>> {
        let (player_type, player_code) = match &ctx.task.details {
            TaskDetails::ApplyEnvironmentalConsequences {
                on_player_type,
                on_player_code,
            } => (on_player_type, on_player_code),
            x => {
                warn!(
                    "ConsequenceEnvironmentHandler called with bad TaskDetail type: {:#?}",
                    x
                );
                return Ok(None);
            }
        };

        let player = match ctx
            .trans
            .find_item_by_type_code(&player_type, &player_code)
            .await?
        {
            None => return Ok(None),
            Some(p) => p,
        };
        let mut player = (*player).clone();

        let mut anything_left = false;
        if player.raddamage > 0 {
            if player.raddamage <= 100 {
                player.raddamage = 0;
            } else {
                anything_left = true;
                if player.raddamage > 2000 {
                    let damage = (player.raddamage - 1000) / 1000;
                    let msg = format!("{} feels unwell", &player.display_for_sentence(1, true));
                    change_health(ctx.trans, -(damage as i64), &mut player, &msg).await?;
                }
                player.raddamage -= 100;
            }
        }
        ctx.trans.save_item_model(&player).await?;
        Ok(if anything_left {
            Some(time::Duration::from_secs(60))
        } else {
            None
        })
    }
}

pub async fn water_environment_effects(
    ctx: &mut TaskRunContext<'_>,
    room: &Room,
    player_item: &mut Item,
) -> DResult<bool> {
    match room.material_type {
        MaterialType::WaterSurface => {}
        MaterialType::Underwater => {
            if change_health(
                &ctx.trans,
                -5,
                player_item,
                &format!(
                    "{} can't breathe underwater",
                    player_item.display_for_sentence(1, true)
                ),
            )
            .await?
            {
                return Ok(true);
            }
        }
        _ => return Ok(false),
    }
    match room.exits.iter().find(|v| v.direction == Direction::DOWN) {
        None => return Ok(false),
        Some(_) => {
            if player_item.active_climb.is_some() {
                return Ok(false);
            }
            if skill_check_and_grind(&ctx.trans, player_item, &SkillType::Swim, 9.0).await? > 0.0 {
                return Ok(true);
            } else {
                broadcast_to_room(
                    &ctx.trans,
                    &player_item.location,
                    None,
                    &format!(
                        "{} seems to be sinking.\n",
                        player_item.display_for_sentence(1, true)
                    ),
                )
                .await?;
                player_item.queue.truncate(0);
                let source = MovementSource::Internal {
                    event_id: Uuid::new_v4(),
                };
                let mut qctx = QueuedCommandContext {
                    trans: ctx.trans,
                    command: &QueueCommand::Movement {
                        direction: Direction::DOWN,
                        source: source.clone(),
                    },
                    item: player_item,
                };
                match attempt_move_immediate(&Direction::DOWN, &mut qctx, &source).await {
                    Ok(_) => {}
                    Err(CommandHandlingError::UserError(_)) => {}
                    Err(CommandHandlingError::SystemError(e)) => return Err(e),
                }
            }
        }
    }
    Ok(true)
}

pub async fn rad_environment_effects(
    ctx: &mut TaskRunContext<'_>,
    room: &Room,
    player_item: &mut Item,
) -> DResult<bool> {
    if room.environment.radiation == 0 {
        return Ok(false);
    }

    let clothes = ctx
        .trans
        .find_by_action_and_location(&player_item.refstr(), &LocationActionType::Worn)
        .await?;
    let shielding: u64 = clothes
        .iter()
        .filter_map(|c| {
            c.possession_type
                .as_ref()
                .and_then(|pt| possession_data().get(&pt))
                .and_then(|pd| pd.wear_data.as_ref())
                .and_then(|wd| wd.soaks.get(&DamageType::Radiation))
                .map(|sd| {
                    (sd.max_soak - sd.min_soak) * rand::thread_rng().gen::<f64>() + sd.min_soak
                })
        })
        .sum::<f64>() as u64;
    if shielding >= room.environment.radiation {
        return Ok(false);
    }

    player_item.raddamage += room.environment.radiation - shielding;
    let existing_task = ctx
        .trans
        .fetch_specific_task("ApplyEnvironmentalConsequences", &player_item.refstr())
        .await?;
    if existing_task.is_none() {
        ctx.trans
            .upsert_task(&Task {
                meta: TaskMeta {
                    task_code: player_item.refstr(),
                    next_scheduled: Utc::now() + chrono::TimeDelta::try_seconds(60).unwrap(),
                    ..Default::default()
                },
                details: TaskDetails::ApplyEnvironmentalConsequences {
                    on_player_type: player_item.item_type.clone(),
                    on_player_code: player_item.item_code.clone(),
                },
            })
            .await?;
    }

    Ok(true)
}

pub async fn passive_health_environment_effects(
    ctx: &mut TaskRunContext<'_>,
    room: &Room,
    player_item: &mut Item,
) -> DResult<bool> {
    if room.environment.passive_health == 0 {
        return Ok(false);
    }

    let msg = format!(
        "{} {}",
        player_item.display_for_sentence(1, true),
        room.environment.passive_health_message
    );
    change_health(
        ctx.trans,
        room.environment.passive_health as i64,
        player_item,
        &msg,
    )
    .await?;

    Ok(true)
}

pub static EFFECT_TASK_HANDLER: &'static (dyn TaskHandler + Sync + Send) =
    &EffectEnvironmentHandler;
pub static CONSEQUENCE_TASK_HANDLER: &'static (dyn TaskHandler + Sync + Send) =
    &ConsequenceEnvironmentHandler;
