use super::{NPCPronounType, NPCSayInfo, NPCSayType, NPCSpawnPossession, NPC};
use crate::{
    models::{
        consent::ConsentType,
        item::{ItemFlag, LocationActionType, Pronouns, SkillType},
    },
    static_content::{
        npc::{npc_pronoun_type_to_pronouns, KillBonus},
        possession_type::PossessionType,
        species::SpeciesType,
    },
};
use serde::Deserialize;
use serde_yaml::from_str as from_yaml_str;

#[derive(Deserialize)]
enum MelbsNPC {
    Citizen {
        code: String,
        name: String,
        spawn_loc: String,
        gender: NPCPronounType,
    },
    Dog {
        code: String,
        adjectives: String,
        spawn_room: String,
    },
    MutantRat {
        code: String,
        adjectives: String,
        spawn_room: String,
    },
    Target {
        code: String,
        name: String,
        spawn_room: String,
        dodge: f64,
    },
}

pub fn npc_list() -> Vec<NPC> {
    use NPCSayType::FromFixedList;
    let melbs_citizen_stdsay = NPCSayInfo {
        say_code: "babble",
        frequency_secs: 120,
        talk_type: FromFixedList(vec![
            "I'm so sick of being cloned.",
            "I hope I don't die again today.",
            "I wish the so-called king would do something about the damned zombies everywhere.",
            "I heard in the olden days before the empire babies grew up naturally instead of being taken away to the body factory.",
            "I know people hated the empire, but I kind of wish it was still intact - it was a lot better than what we have now.",
            "I wish there wasn't so much radiation outside of Melbs CBD.",
            "I heard about a guy who went to a special place somewhere around here, and there was a machine that enhanced his wristpad and gave him basically superpowers.",
            "The damn vampire movement... they are all so sneaky, and I never know when they are going to come for my blood.",
        ])
    };

    from_yaml_str::<Vec<MelbsNPC>>(include_str!("melbs_npcs.yaml"))
        .unwrap()
        .into_iter()
        .map(|c|
             match c {
                 MelbsNPC::Citizen {
                     code, name, gender, spawn_loc
                 } =>
                     NPC {
                         code: format!("melbs_citizen_{}", &code),
                         name,
                         pronouns: npc_pronoun_type_to_pronouns(&gender),
                         description: "A fairly ordinary looking citizen of Melbs, clearly weary from the harsh reality of post-apocalyptic life".to_owned(),
                         spawn_location: format!("room/melbs_{}", &spawn_loc),
                         spawn_possessions: vec![
                             NPCSpawnPossession {
                                 what: PossessionType::Shirt,
                                 action_type: LocationActionType::Worn,
                                 wear_layer: 0,
                             },
                             NPCSpawnPossession {
                                 what: PossessionType::Jeans,
                                 action_type: LocationActionType::Worn,
                                 wear_layer: 0,
                             },
                         ],
                         message_handler: None,
                         wander_zones: vec!("melbs".to_owned()),
                         says: vec!(melbs_citizen_stdsay.clone()),
                         player_consents: vec!(ConsentType::Medicine, ConsentType::Share),
                         ..Default::default()
                     },
                 MelbsNPC::Dog { code, adjectives, spawn_room } =>
                     NPC {
                         code: format!("melbs_dog_{}", &code),
                         name: format!("{} dog", &adjectives),
                         pronouns: Pronouns { is_proper: false, ..Pronouns::default_inanimate() },
                         aggression: 12,
                         wander_zones: vec!("melbs".to_owned()),
                         description: "A malnourished looking dog. Its skeleton is visible through its thin and patchy fur. It smells terrible, and certainly doesn't look tame.".to_owned(),
                         aliases: vec!("dog".to_owned()),
                         spawn_location: format!("room/{}", spawn_room),
                         intrinsic_weapon: Some(PossessionType::Fangs),
                         species: SpeciesType::Dog,
                         kill_bonus: Some(KillBonus {
                             msg: "On your wristpad: Thank you for helping Melbs with animal control! Here's your fee.",
                             payment: 100,
                         }),
                         player_consents: vec!(ConsentType::Fight),
                         ..Default::default()
                     },
                 MelbsNPC::MutantRat { code, adjectives, spawn_room } =>
                     NPC {
                         code: format!("melbs_rat_{}", &code),
                         name: format!("{} mutant rat", &adjectives),
                         pronouns: Pronouns { is_proper: false, ..Pronouns::default_inanimate() },
                         aggression: 21,
                         wander_zones: vec!("kings_office".to_owned()),
                         description: "Something that resembles a brown rat - except it is about 2 metres tall! Huge teeth poke through its lips, and its feet betray brutal looking claws. Gigantic whiskers poke from its face".to_owned(),
                         aliases: vec!("rat".to_owned()),
                         spawn_location: format!("room/{}", spawn_room),
                         intrinsic_weapon: Some(PossessionType::Fangs),
                         species: SpeciesType::Rat,
                         aggro_pc_only: true,
                         kill_bonus: Some(KillBonus {
                             msg: "On your wristpad: Thank you for helping Melbs with animal control! Here's your fee.",
                             payment: 150,
                         }),
                         total_skills: SkillType::values()
                             .into_iter()
                             .map(|sk| {
                                 (
                                     sk.clone(),
                                     match sk {
                                         SkillType::Dodge => 15.0,
                                         SkillType::Fists => 15.0,
                                         _ => 8.0
                                     }
                                 )
                             }).collect(),
                         player_consents: vec!(ConsentType::Fight),
                         ..Default::default()
                     },
                 MelbsNPC::Target { code, name, spawn_room, dodge } => {
                     NPC {
                         code: format!("melbs_target_{}", &code),
                         name: name.clone(),
                         pronouns: Pronouns { is_proper: true, ..Pronouns::default_inanimate() },
                         description: "A robotic device that seems to be designed for target practice. It includes a stand, and has the vague shape of a robot torso. It seems designed to take a lot of damage, and to move to dodge bullets".to_owned(),
                         aliases: vec!("target".to_owned()),
                         spawn_location: format!("room/{}", spawn_room),
                         species: SpeciesType::Robot,
                         max_health: 10000,
                         extra_flags: vec![ItemFlag::DontCounterattack],
                         total_skills: SkillType::values()
                             .into_iter()
                             .map(|sk| {
                                 (
                                     sk.clone(),
                                     match sk {
                                         SkillType::Dodge => dodge,
                                         _ => 8.0
                                     }
                                 )
                             }).collect(),
                         player_consents: vec!(ConsentType::Fight),
                         ..Default::default()
                     }
                 }
             },
        ).collect()
}
