use crate::{
    message_handler::user_commands::{user_error, UResult, VerbContext},
    models::{item::Item, journal::JournalType},
    static_content::{journals::award_journal_if_needed, room::ScanCode},
};
use ansi::ansi;
use async_trait::async_trait;
use once_cell::sync::OnceCell;

use super::{PossessionData, PossessionType, WriteHandler};

pub struct RadTestHandler {}

#[async_trait]
impl WriteHandler for RadTestHandler {
    async fn write_cmd(
        &self,
        ctx: &mut VerbContext,
        player: &Item,
        on_what: &Item,
        write_what: &str,
    ) -> UResult<()> {
        let input = write_what.trim().to_uppercase();
        if player.location != "room/oorans_testing" {
            user_error(
                "You realise you are probably supposed to do this in the testing centre."
                    .to_owned(),
            )?;
        }
        if input.len() != 4 {
            user_error(
                "You realise you'll need to write 4 letters to complete the test.".to_owned(),
            )?;
        }
        if input.chars().any(|x| !"ABCD".contains(x)) {
            user_error("You realise you can only pick A, B, C or D for each answer.".to_owned())?;
        }
        let wrong = input
            .chars()
            .zip("BCAC".chars())
            .filter(|(x, y)| x != y)
            .count();

        if wrong > 0 {
            ctx.trans
                .queue_for_session(
                    ctx.session,
                    Some(&format!(
                        "The pen makes a scratching sound as you mark the paper with it. An invigilator passes by, sees you have \
                         completed your test, and shakes her head. \"You got {} wrong!\", she says, \"Radiation safety is very \
                         important, so we'll have to fail you. Try visting the OORANS Training Centre before re-attempting the \
                         test\"! She takes the spoiled test off you.\n",
                        wrong
                    )),
                )
                .await?;
        } else {
            ctx.trans.queue_for_session(ctx.session,
                                        Some("The pen makes a scratching sound as you mark the paper with it. An invigilator passes by, sees you have \
                                         completed your test, and smiles at you. \"Congratulations\", she says, \"You got 100%, and so are eligible \
                                         for a permit to travel to radiation contaminated areas! Be safe out there\"! She takes the completed test off you, and holds her wristpad up to yours, causing yours to beep and show a copy of the permit.\n")

            ).await?;
            if let Some(user) = ctx.user_dat.as_mut() {
                let mut player = (*player).clone();
                let awarded_journal = award_journal_if_needed(
                    &ctx.trans,
                    user,
                    &mut player,
                    JournalType::GotRadPermit,
                )
                .await?;
                if awarded_journal {
                    ctx.trans.save_item_model(&player).await?;
                }
                let added_permit = if !user.scan_codes.contains(&ScanCode::RadPermit) {
                    user.scan_codes.push(ScanCode::RadPermit);
                    true
                } else {
                    false
                };
                if awarded_journal || added_permit {
                    ctx.trans.save_user_model(user).await?;
                }
            }
        }
        ctx.trans
            .delete_item(&on_what.item_type, &on_what.item_code)
            .await?;
        Ok(())
    }
}

static RADTEST_HANDLER: RadTestHandler = RadTestHandler {};

pub fn data() -> &'static Vec<(PossessionType, PossessionData)> {
    static D: OnceCell<Vec<(PossessionType, PossessionData)>> = OnceCell::new();
    &D.get_or_init(|| vec!(
        (PossessionType::RadSafetyTest,
         PossessionData {
             display: "a radiation safety test paper",
             details: ansi!("A test paper that looks like it is for a multiple choice test. The heading reads \"Radiation Safety Licence Test\". Bold text below it says: \"Test-takers beware! To ensure academic integrity, it is not possible to erase your answers once written - you'll only get one attempt and need to buy a new paper if you spoil it. Write your answers using the letter of your choice, in all caps, with no spaces. For example, to answer A for question 1, B for question 2, C for question 3, and D for question 4, you would <bold>write ABCD on test<reset>.\n\n\
                             Question 1: Suppose a person was to spend a long time walking in a heavily contaminated radfield with no special equipment. What would happen?\n\
                             A: Nothing would happen, as long as they didn't eat the contaminated dirt.\n\
                             B: They would not immediately feel any ill effects, but would eventually succomb to radiation poisoning.\n\
                             C: They would immediately feel a burning sensation from the ionising radiation, due to the area being radiologically hot.\n\
                             D: They would be able to tell the area is contaminated because it gives off a green glow.\n\
                             Question 2: What device can measure how much ionising radiation is in an area?\n\
                             A: A Dowsing Rod.\n\
                             B: A Colourimetric Radiation Detector.\n\
                             C: A Geiger Counter.\n\
                             D: A Schottky Power Detector.\n\
                             Question 3: How can you reduce your ionising radiation exposure while travelling in a contaminated area?\n\
                             A: Wear an OORANS certified radiation suit.\n\
                             B: Use a safety shower immediately after travelling.\n\
                             C: Eat a small amount of fallout dust daily to build up resistance.\n\
                             D: Move back and forwards within the contaminated area to ensure radiation misses you.\n\
                             Question 4: What should you do if you are accidentally exposed to ionising radiation?\n\
                             A: Walk backwards through the area you were exposed to reverse the effect.\n\
                             B: Go to the Melbs General Hospital.\n\
                             C: Find someone skilled as a medic and ask them to use a rad detox kit on you.\n\
                             D: Nothing, radiation isn't dangerous"),
             aliases: vec!("test paper"),
             weight: 10,
             write_handler: Some(&RADTEST_HANDLER),
             ..Default::default()
         }),
    ))
}
