use super::{Room, RoomSellTrigger, ScanCode, SimpleRoom};
use crate::{
    message_handler::user_commands::{UResult, UserError, VerbContext},
    models::{
        item::{Item, Scavtype},
        journal::JournalType,
        user::QuestProgress,
    },
    static_content::{
        journals::award_journal_if_needed, possession_type::PossessionType, scavtable::Scavinfo,
    },
};
use ansi::ansi;
use async_trait::async_trait;
use serde_yaml::from_str as from_yaml_str;

pub fn sewer_scavtable() -> Vec<Scavinfo> {
    vec![Scavinfo {
        possession_type: PossessionType::RustySpike,
        p_present: 1.0,
        difficulty_mean: 7.0,
        difficulty_stdev: 1.0,
        scavtype: Scavtype::Scavenge,
    }]
}

pub struct JosephineSellTrigger;

#[async_trait]
impl RoomSellTrigger for JosephineSellTrigger {
    async fn handle_sell(
        &self,
        ctx: &mut VerbContext,
        _room: &Room,
        player_item: &Item,
        sell_item: &Item,
    ) -> UResult<()> {
        if sell_item.possession_type == Some(PossessionType::RadiantPredatorDagger) {
            let user_dat = ctx
                .user_dat
                .as_mut()
                .ok_or_else(|| UserError("Selling while not logged in?".to_owned()))?;
            if user_dat.scan_codes.contains(&ScanCode::RedImperialCode) {
                return Ok(());
            }
            match user_dat.quest_progress.as_mut() {
                None => {
                    user_dat.quest_progress = Some(QuestProgress {
                        daggers_sold_to_josephine: 1,
                        ..Default::default()
                    });
                }
                Some(qd) => {
                    qd.daggers_sold_to_josephine += 1;
                }
            }
            if user_dat
                .quest_progress
                .as_ref()
                .map(|qp| qp.daggers_sold_to_josephine)
                .unwrap_or(0)
                >= 5
            {
                user_dat.scan_codes.push(ScanCode::RedImperialCode);
                ctx.trans.queue_for_session(
                    &ctx.session,
                    Some(ansi!("<blue>Josephine whispers to you: \"Thank you - those daggers will really help me \
                                build my defence system.\"<reset>\n\
                                Josephine takes your hand and enters a code onto your wristpad. It beeps and flashes up \
                                a message saying: \"Red Imperial Code saved\".\n"))
                ).await?;
                let mut player_item = (*player_item).clone();
                award_journal_if_needed(
                    &ctx.trans,
                    user_dat,
                    &mut player_item,
                    JournalType::BribedJosephineForRedCode,
                )
                .await?;
            }
            ctx.trans.save_user_model(user_dat).await?;
            return Ok(());
        }
        Ok(())
    }
}

pub fn room_list() -> Vec<Room> {
    from_yaml_str::<Vec<SimpleRoom<()>>>(include_str!("melbs_sewers.yaml"))
        .unwrap()
        .into_iter()
        .map(|r| r.into())
        .map(|r: Room| {
            if r.code == "melbs_sewers_subsewer_josephine" {
                Room {
                    sell_trigger: Some(Box::new(JosephineSellTrigger)),
                    ..r
                }
            } else {
                r
            }
        })
        .collect()
}
